//---------------------------------------------------------------------------
//  Mpcx.pas
//  A "Graphics" unit's extention for PCX image format.
//  Copyright ( c ) 2000, 01 by Mental Studio - http://mentals.126.com
//  Author : Raptor - raptorz@163.com
//---------------------------------------------------------------------------
unit Mpcx;

interface

Uses
	Classes, Graphics;

Type
	TMPCXImage = Class ( TBitmap )
    Public
    	Procedure LoadFromStream( aStream : TStream ); Override;
        Procedure SaveToStream(   aStream : TStream ); Override;
    End;

implementation

{$L MpcxData.obj}

Function  MPCX_Size(  aMPcxData : Pointer; aSize : Integer ) : Integer; External;
Procedure MPCX_Read(  aMPcxData : Pointer; aSize : Integer; aBmData   : Pointer ); External;
Function  MPCX_Write( aBmData   : Pointer; aSize : Integer; aMPcxData : Pointer ) : Integer; External;

//  External functions
Function _memalloc( aSize : Integer ) : Pointer; cdecl;
Begin
	GetMem( Result, aSize );
End;

Procedure _memfree( aMem : Pointer ); cdecl;
Begin
	FreeMem( aMem );
End;

//  This is a C function but not Delphi
Function _memset( s : Pointer; c : Byte; n : Integer ) : Pointer; cdecl;
Begin
	FillChar( s^, n, c );
    Result := s;
End;

//  TMPCXImage

Procedure TMPCXImage.LoadFromStream( aStream : TStream );
Var
	pPcx : TMemoryStream;
    pBmp : TMemoryStream;
Begin
	If ( aStream Is TMemoryStream ) Then
    	pPcx := aStream As TMemoryStream
    Else
    Begin
    	pPcx := TMemoryStream.Create;
        pPcx.CopyFrom( aStream, aStream.Size );
    End;
    pBmp := TMemoryStream.Create;
    Try
    	pBmp.Size := MPCX_Size( pPcx.Memory, pPcx.Size );
        If ( pBmp.Size > 0 ) Then
        Begin
        	MPCX_Read( pPcx.Memory, pPcx.Size, pBmp.Memory );
            Inherited LoadFromStream( pBmp );
        End;
    Finally
    	pBmp.Free;
        If ( Not ( aStream Is TMemoryStream ) ) Then
        	pPcx.Free;
    End;
End;

Procedure TMPCXImage.SaveToStream( aStream : TStream );
Var
	n : Integer;
    pPcx : TMemoryStream;
    pBmp : TMemoryStream;
Begin
	pPcx := TMemoryStream.Create;
    pBmp := TMemoryStream.Create;
    Try
    	Inherited SaveToStream( pBmp );
        pPcx.Size := Round( pBmp.Size * 1.5 );  //  size * 1.5
        n := MPCX_Write( pBmp.Memory, pBmp.Size, pPcx.Memory );
        If ( n > 0 ) Then
        Begin
        	pPcx.Size := n;
            aStream.CopyFrom( pPcx, n );
        End;
    Finally
    	pBmp.Free;
        pPcx.Free;
    End;
End;

//initialization
//	TPicture.RegisterFileFormat( 'pcx', 'PCX Image Files', TMPCXImage );

//finalization
//	TPicture.UnregisterGraphicClass( TMPCXImage );

end.

