//---------------------------------------------------------------------------
//  MpcxData.c
//  Converting functions between PCX Image data and BMP Image data,
//    It's used by unit MPcx.
//  Copyright ( c ) 2000, 01 by Mental Studio - http://mentals.126.com
//  Author : Raptor - raptorz@163.com
//---------------------------------------------------------------------------
//
/*
	Ԫñ׼ C дɣԡ
    
     DELPHI е÷ļMpcx.pas
     BCB    е÷ļMpcx.h/Mpcx.cpp

    Ҫ Delphi 뱾Ԫ MpcxData.obj ļ
    bcc32 -c -pr -O2 -C -K -N- -k- -d -3 -r- MpcxData.c
    ֱΪ
    -c  Ϊ obj ļ
    -pr  Pascal  fastcall ÷ʽ
    -O2 ŻѡΪ2
    -C  Ƕע
    -K  Ĭʹ޷ַ
    -N- ջ
    -k- ʹñ׼ջ
    -d  ϲظַ
    -3  ʹ 386 ָ
    -r- δ֪^_^
*/

#include <Windows.h>
#include "MpcxData.h"

//---------------------------------------------------------------------------

typedef unsigned char  BYTE;
typedef unsigned short WORD;
typedef unsigned char *PBYTE;

//---------------------------------------------------------------------------

typedef struct {
	BYTE manufacturer; // 0x0a
	BYTE version;      // 0x05
	BYTE encoding;     // 0x01
	BYTE bits_per_pixel;
	WORD xmin,ymin;
	WORD xmax,ymax;
	WORD hres;
	WORD vres;
	BYTE palette[48];
	BYTE reserved;
	BYTE color_planes;
	WORD bytes_per_line;
	WORD palette_type;
	BYTE filler[58];
} PCXHEAD, *LPPCXHEAD;

//---------------------------------------------------------------------------
//  Unpack a PCX line
//  Used only this file
//---------------------------------------------------------------------------
static int GetPCXLine( PBYTE const aDest, const BYTE * const aSrc, int aCount )
{
	BYTE        c, i;
    const BYTE *pSrc = aSrc;
    BYTE       *pDest = aDest;

	do {
    	c = *pSrc++;  // get a key byte
		if ( ( c & 0xc0 ) == 0xc0 ) // it's a run of bytes field
		{
			i = ( BYTE )( c & 0x3f ); // i: times to store
			c = *pSrc++;              // c: which to store
			while ( i-- )
            	*pDest++ = c;
		}
		else  // not a run of bytes field,just store it.
        	*pDest++ = c;
	} while ( ( pDest - aDest ) < aCount );

    return pSrc - aSrc;  // Return unpacked source length
}

//---------------------------------------------------------------------------
//  Pack a PCX line
//  Used only this file
//---------------------------------------------------------------------------
static int PutPCXLine( PBYTE const aDest, const BYTE * const aSrc, int aCount )
{
	BYTE        c, i;
    const BYTE *pSrc = aSrc;
    BYTE       *pDest = aDest;

	do {
		i = 1;
        c = *pSrc++;
		while ( ( *pSrc == c ) && ( ( pSrc - aSrc ) < aCount ) && ( i < 63 ) )
        { // Count run length
        	i++;
			pSrc++;
        }
		if ( i > 1 )
		{ // run length encode
			*pDest++ = ( BYTE )( i | 0xc0 );
            *pDest++ = c;
		}
		else
		{ // alone char
			if ( ( c & 0xc0 ) == 0xc0 )
            	*pDest++ = 0xc1;
			*pDest++ = c;
		}
	} while ( ( pSrc - aSrc ) < aCount );

    return pDest - aDest;  // Return packed line length
}

//---------------------------------------------------------------------------
//  Return BMP data size
//---------------------------------------------------------------------------
int MPCX_Size( const void * const aMPcxData, int aSize )
{
    int nBits, nWidth, nHeight, nDataLen, nPalLen;

	LPPCXHEAD pHead = ( LPPCXHEAD )aMPcxData;

    if ( aSize <= sizeof ( PCXHEAD ) )
    	return 0;

    nBits   = pHead->bits_per_pixel * pHead->color_planes; // bits per pixel
	nWidth  = pHead->xmax - pHead->xmin + 1;
    nHeight = pHead->ymax - pHead->ymin + 1;
    nDataLen= ( ( ( nWidth * nBits + 7 ) / 8 + 3 ) & 0xFFFFFFFC ) * nHeight; // image size by byte, align by 32bit
    nPalLen = ( nBits <= 8 ) ? ( ( 1L << nBits ) * sizeof ( RGBQUAD ) ) : 0; // palette size, 24bit without palette

    return ( sizeof ( BITMAPFILEHEADER ) + sizeof ( BITMAPINFOHEADER ) + nPalLen + nDataLen ); // total BMP size;
}
//---------------------------------------------------------------------------
//  Convert PCX data to BMP data
//---------------------------------------------------------------------------
void MPCX_Read( const void * const aMPcxData, int aSize, void * const aBmData )
{
	int  i, j;
	LPPCXHEAD pHead = ( LPPCXHEAD )aMPcxData;
    int  nBits   = pHead->bits_per_pixel * pHead->color_planes;
	int  nWidth  = pHead->xmax - pHead->xmin + 1;
    int  nHeight = pHead->ymax - pHead->ymin + 1;
//    int  nPLen   = ( ( nWidth * pHead->bits_per_pixel + 7 ) / 8 + 1 ) & 0xFFFFFFFE; //  Bytes per line and plane ( PCX ), align by WORD;
	int  nPLen   = pHead->bytes_per_line;
    int  nBLen   = ( ( nWidth * nBits + 7 ) / 8 + 3 ) & 0xFFFFFFFC; //  align by DWORD

    LPBITMAPFILEHEADER pBfh = ( LPBITMAPFILEHEADER )aBmData;
    LPBITMAPINFOHEADER pBih = ( LPBITMAPINFOHEADER )( ( BYTE * )aBmData + sizeof ( BITMAPFILEHEADER ) );
    LPRGBQUAD          pPal = ( LPRGBQUAD )( ( BYTE * )pBih + sizeof ( BITMAPINFOHEADER ) );
    BYTE             * pData= ( BYTE * )pPal + ( ( nBits <= 8 ) ? ( ( 1L << nBits ) * sizeof ( RGBQUAD ) ) : 0 ); //  add palette size

    BYTE * p;
    BYTE * pSrc;
    BYTE * pDest;
    BYTE * pLine;
    BYTE b40, b41, b42, b43; // Only used in 4bit color

    if ( ( pHead->manufacturer != 0x0a ) || ( pHead->version != 0x05 ) || ( pHead->encoding != 0x01 ) )
    	return;

    pBfh->bfType      = 'BM';
    pBfh->bfSize      = pData - ( BYTE * )pBfh + nBLen * nHeight; // total size
    pBfh->bfReserved1 = 0;
    pBfh->bfReserved2 = 0;
    pBfh->bfOffBits   = pData - ( BYTE * )pBfh;

    pBih->biSize          = 40;
	pBih->biWidth         = nWidth;
	pBih->biHeight        = nHeight;
	pBih->biPlanes        = 1;
	pBih->biBitCount      = nBits;
	pBih->biCompression   = BI_RGB;
	pBih->biSizeImage     = pBfh->bfSize - pBfh->bfOffBits;
	pBih->biXPelsPerMeter = 0;
	pBih->biYPelsPerMeter = 0;
	pBih->biClrUsed       = 0;
	pBih->biClrImportant  = 0;

    if ( nBits <= 8 )
    {
    	//  palette pointer, Caution : [-769] == 0x0C
        if ( nBits < 8 )
        	p = &pHead->palette[0];
        else
        	p = ( BYTE * )aMPcxData + aSize - 768;
		// Read palette
    	for ( i = 0; i < ( ( int )1 << nBits ); i++ )
        {
        	pPal[i].rgbRed   = *p++;
            pPal[i].rgbGreen = *p++;
            pPal[i].rgbBlue  = *p++;
        }
    } //  24bit image without palette

    pSrc = ( BYTE * )aMPcxData + sizeof ( PCXHEAD ); // start of PCX packed data
    pDest = pData + nBLen * ( nHeight - 1 );
    if ( pHead->color_planes == 1 ) // 1bit / 8bit
    {
	    for ( i = 0; i < nHeight; i++ )
    	{  //  Read a lines
    		pSrc += GetPCXLine( pDest, pSrc, nPLen );
            pDest -= nBLen;
	    }
    }
    else
    {
    	pLine = ( BYTE * )memalloc( nPLen * pHead->color_planes );  //  line buffer
    	if ( pHead->color_planes == 3 )  //  24bit
	    {
		    for ( i = 0; i < nHeight; i++ )
    		{  //  Read 3 lines
    			for ( j = 0; j < 3; j++ )
					pSrc += GetPCXLine( pLine + j * nPLen, pSrc, nPLen );
                p = pDest;
                for ( j = 0; j < nPLen; j++ )
                {
                	*p++ = pLine[j + nPLen * 2];
                    *p++ = pLine[j + nPLen];
                    *p++ = pLine[j];
                }
                pDest -= nBLen;
		    }
    	}
	    else if ( pHead->color_planes == 4 ) // 4bit
    	{
	    	for ( i = 0; i < nHeight; i++ )
	    	{  //  Read 4 lines
    			for ( j = 0; j < 4; j++ )
 					pSrc += GetPCXLine( pLine + j * nPLen, pSrc, nPLen );
                p = pDest;
                for ( j = 0; j < nPLen; j++ )
                { // Can you understand these codes ? ^_^
                	b40 = pLine[j];
                    b41 = pLine[j + nPLen];
                    b42 = pLine[j + nPLen * 2];
                    b43 = pLine[j + nPLen * 3];
                	*p++ =( ( b40 & 0x80 ) >> 3 ) | ( ( b41 & 0x80 ) >> 2 )
                    	| ( ( b42 & 0x80 ) >> 1 ) |   ( b43 & 0x80 )
                        | ( ( b40 & 0x40 ) >> 6 ) | ( ( b41 & 0x40 ) >> 5 )
                        | ( ( b42 & 0x40 ) >> 4 ) | ( ( b43 & 0x40 ) >> 3 );
                	*p++ =( ( b40 & 0x20 ) >> 1 ) |   ( b41 & 0x20 )
                    	| ( ( b42 & 0x20 ) << 1 ) | ( ( b43 & 0x20 ) << 2 )
                        | ( ( b40 & 0x10 ) >> 4 ) | ( ( b41 & 0x10 ) >> 3 )
                        | ( ( b42 & 0x10 ) >> 2 ) | ( ( b43 & 0x10 ) >> 1 );
                	*p++ =( ( b40 & 0x08 ) << 1 ) | ( ( b41 & 0x08 ) << 2 )
                    	| ( ( b42 & 0x08 ) << 3 ) | ( ( b43 & 0x08 ) << 4 )
                        | ( ( b40 & 0x04 ) >> 2 ) | ( ( b41 & 0x04 ) >> 1 )
                        |   ( b42 & 0x04 )        | ( ( b43 & 0x04 ) << 1 );
                	*p++ =( ( b40 & 0x02 ) << 3 ) | ( ( b41 & 0x02 ) << 4 )
                    	| ( ( b42 & 0x02 ) << 5 ) | ( ( b43 & 0x02 ) << 6 )
                        |   ( b40 & 0x01 )        | ( ( b41 & 0x01 ) << 1 )
                        | ( ( b42 & 0x01 ) << 2 ) | ( ( b43 & 0x01 ) << 3 );
                }
                pDest -= nBLen;
		    }
    	}
        memfree( pLine );
    }
}
//---------------------------------------------------------------------------
//  Convert BMP data to PCX data and return PCX data size
//---------------------------------------------------------------------------
int MPCX_Write( const void * const aBmData, int aSize, void * const aMPcxData )
{
	int  i, j;

    LPBITMAPINFOHEADER pBih = ( LPBITMAPINFOHEADER )( ( BYTE * )aBmData + sizeof ( BITMAPFILEHEADER ) );
    LPRGBQUAD          pPal = ( LPRGBQUAD )( ( BYTE * )pBih + sizeof ( BITMAPINFOHEADER ) );

    int  nBits   = pBih->biBitCount;
	int  nWidth  = pBih->biWidth;
    int  nHeight = pBih->biHeight;
    int  nBLen   = ( ( nWidth * nBits + 7 ) / 8 + 3 ) & 0xFFFFFFFC; //  align by DWORD
    int  nPLen;

    BYTE * pData= ( BYTE * )pPal + ( ( nBits <= 8 ) ? ( ( 1L << nBits ) * sizeof ( RGBQUAD ) ) : 0 ); //  add palette size

	LPPCXHEAD pHead = ( LPPCXHEAD )aMPcxData;

    BYTE * p;
    BYTE * pSrc;
    BYTE * pDest;
    BYTE * pLine;
    BYTE b40, b41, b42, b43; // Only used in 4bit color

    //  init head
    memset( pHead, 0, sizeof ( PCXHEAD ) );
    pHead->manufacturer = 0x0a;
    pHead->version      = 0x05;
	pHead->encoding     = 0x01;
    switch ( nBits ) {
    case 4 :
    	pHead->bits_per_pixel = 1;
        pHead->color_planes   = 4;
        break;
    case 24 :
    	pHead->bits_per_pixel = 8;
        pHead->color_planes   = 3;
        break;
    default :
    	pHead->bits_per_pixel = nBits;
        pHead->color_planes   = 1;
        break;
    }

    nPLen = ( ( nWidth * pHead->bits_per_pixel + 7 ) / 8 + 1 ) & 0xFFFFFFFE; //  Bytes per line and plane ( PCX ), align by WORD;;

	pHead->xmax = nWidth - 1;
	pHead->ymax = nHeight - 1;
	pHead->bytes_per_line = nPLen;
	pHead->palette_type   = 0x01;

    if ( nBits <= 4 )
    {
       	p = &pHead->palette[0];
		// Read palette
    	for ( i = 0; i < ( ( int )1 << nBits ); i++ ) // read palette
        {
        	*p++ = pPal[i].rgbRed;
            *p++ = pPal[i].rgbGreen;
            *p++ = pPal[i].rgbBlue;
        }
    } //  8bit palette at the end of data, 24bit image without palette

    pDest = ( BYTE * )aMPcxData + sizeof ( PCXHEAD ); // start of PCX packed data
    pSrc = pData + nBLen * ( nHeight - 1 );
    if ( pHead->color_planes == 1 ) // 1bit / 8bit
    {
	    for ( i = 0; i < nHeight; i++ )
    	{  //  Write a line
        	pDest += PutPCXLine( pDest, pSrc, nPLen );
            pSrc -= nBLen;
	    }

	    //  8bit palette at the end of data
    	if ( nBits == 8 )
	    {
    		p = pDest;
        	*p++ = 0x0c; // palette flag
	    	for ( i = 0; i < 256; i++ ) // write palette
    	    {
        		*p++ = pPal[i].rgbRed;
            	*p++ = pPal[i].rgbGreen;
	            *p++ = pPal[i].rgbBlue;
    	    }
        	pDest = p;
	    }
    }
    else
    {
    	pLine = ( BYTE * )memalloc( nPLen * pHead->color_planes );  //  line buffer
    	if ( pHead->color_planes == 3 ) // 24bit
	    {
		    for ( i = 0; i < nHeight; i++ )
    		{
                p = pSrc;
                for ( j = 0; j < nPLen; j++ ) // split color planes
                {
                	pLine[j + nPLen * 2] = *p++;
                    pLine[j + nPLen] = *p++;
                    pLine[j] = *p++;
                }
    			for ( j = 0; j < 3; j++ ) // pack 3 lines
					pDest += PutPCXLine( pDest, pLine + j * nPLen, nPLen );
                pSrc -= nBLen;
		    }
    	}
	    else if ( pHead->color_planes == 4 ) // 4bit
    	{
	    	for ( i = 0; i < nHeight; i++ )
	    	{
                p = pSrc;
                for ( j = 0; j < nPLen; j++ )
                {
                	b40 = *p++;
                    b41 = *p++;
                    b42 = *p++;
                    b43 = *p++;
                	pLine[j]
                    	= ( ( b40 & 0x10 ) << 3 ) | ( ( b40 & 0x01 ) << 6 )
                    	| ( ( b41 & 0x10 ) << 1 ) | ( ( b41 & 0x01 ) << 4 )
                        | ( ( b42 & 0x10 ) >> 1 ) | ( ( b42 & 0x01 ) << 2 )
                        | ( ( b43 & 0x10 ) >> 3 ) |   ( b43 & 0x01 );
                    pLine[j + nPLen]
                    	= ( ( b40 & 0x20 ) << 2 ) | ( ( b40 & 0x02 ) << 5 )
                        |   ( b41 & 0x20 )        | ( ( b41 & 0x02 ) << 3 )
                        | ( ( b42 & 0x20 ) >> 2 ) | ( ( b42 & 0x02 ) << 1 )
                        | ( ( b43 & 0x20 ) >> 4 ) | ( ( b42 & 0x02 ) >> 1 );
                    pLine[j + nPLen * 2]
                    	= ( ( b40 & 0x40 ) << 1 ) | ( ( b40 & 0x04 ) << 4 )
                        | ( ( b41 & 0x40 ) >> 1 ) | ( ( b41 & 0x04 ) << 2 )
                        | ( ( b42 & 0x40 ) >> 3 ) |   ( b42 & 0x04 )
                        | ( ( b43 & 0x40 ) >> 5 ) | ( ( b43 & 0x04 ) >> 2 );
                    pLine[j + nPLen * 3]
                    	=   ( b40 & 0x80 )        | ( ( b40 & 0x08 ) << 3 )
                        | ( ( b41 & 0x80 ) >> 2 ) | ( ( b41 & 0x08 ) << 1 )
                        | ( ( b42 & 0x80 ) >> 4 ) | ( ( b42 & 0x08 ) >> 1 )
                        | ( ( b43 & 0x80 ) >> 6 ) | ( ( b43 & 0x08 ) >> 3 );
                }
    			for ( j = 0; j < 4; j++ )
 					pDest += PutPCXLine( pDest, pLine + j * nPLen, nPLen );
                pSrc -= nBLen;
		    }
    	}
        memfree( pLine );
    }
	return ( pDest - ( BYTE * )aMPcxData );
}
//---------------------------------------------------------------------------
 
